# Copyright: Tian Chen and Michael Monagan, 2024

CCodeDir := "./"; 

DetBareissSymm := define_external('detBareissSymmetric', 
  bb::ARRAY(0..nn,datatype=integer[8]),
  nn::integer[4],
  pp::integer[8],
  LIB=cat(CCodeDir, "detinterp3.so"),
  RETURN::integer[8]):
DetBareissSymm := subsop(1=(bb,nn,pp),eval(DetBareissSymm)):

CallBB := define_external('callBB', 
  bb::ARRAY(1..nn,datatype=integer[8]), # Eval pts, indexed from 1 to N (Maple index)
  nn::integer[4],                       # N, no.of variables or eval pts
  vii::integer[4], vjj::integer[4],     # 1 <= xi < xj <= N (Maple index)
  dii::integer[8], djj::integer[8],     # di and dj
  XX::ARRAY(0..nn,datatype=integer[8]), # size dmax+1, where dmax=max(di,dj)
  MM::ARRAY(0..dii,0..djj,datatype=integer[8]), # size(dj+1)*(di+1)
  pp::integer[8],
  cnt::integer[8],
  LIB=cat(CCodeDir, "detinterp3.so"),
  RETURN::integer[8]):

Det64s:= define_external('det64s',
  AA::ARRAY(1..nn,1..nn,integer[8],order=C_order), 
  nn::integer[4], 
  pp::integer[8], 
  LIB=cat(CCodeDir,"detinterp3.so"), 
  RETURN::integer[8]):
Det64s := subsop(1=(AA,nn,pp),eval(Det64s)):

#evallistlist := define_external('evalLIST64s',
#    LL::integer[8],
#    XX::integer[8],
#    aa::ARRAY(1..nn,integer[8]),
#    AA::ARRAY(1..nn,1..nn,integer[8]),
#    pp::integer[8],
#    RETURN::integer[8],
#    LIB=cat(CCodeDir,"gcd8.so") ): 
#evallistlist := subsop(1=(LL,XX,aa,AA,pp),eval(evallistlist)):

NewtonInv := define_external('NewtonInverses', 
  aa::ARRAY(0..nn,datatype=integer[8]), #size n, indexed from 0 to n-1
  nn::integer[8],                       #size n, deg n-1
  II::ARRAY(0..nn,datatype=integer[8]), #size n, indexed from 0 to n-1
  pp::integer[8],
  LIB=cat(CCodeDir, "detinterp3.so") ):

Interp2var := define_external('interp2var2',
  vii::integer[4], vjj::integer[4],     # 1 <= xi < xj <= N (Maple index)
  dii::integer[8], djj::integer[8], dmm::integer[8], # dmax = max(di,dj)
  XX::ARRAY(0..nn,datatype=integer[8]), # size dmax+1, where dmax=max(di,dj) 
  II::ARRAY(0..nn,datatype=integer[8]), # size dmax+1
  MM::ARRAY(0..dii,0..djj,datatype=integer[8]), # size(dj+1)*(di+1)
  pp::integer[8],
  LIB=cat(CCodeDir, "detinterp3.so") ):

getmapleID := define_external('getMapleID',
   aa::integer[8],
   LIB=cat(CCodeDir, "gcd5.so"),
   RETURN::integer[4]):

getsupport := define_external('getsupport64s',
  ff::integer[8],
  nf::integer[4],
  ALPHA::ARRAY(1..nn,integer[8]),
  nn::integer[4],
  xx::integer[4],
  dx::integer[4],
  TT::ARRAY(0..nn,integer[4]),
  CC::ARRAY(1..nn,integer[8]),
  MM::ARRAY(1..nn,integer[8]),
  pp::integer[8],
  LIB=cat(CCodeDir, "gcd5.so") ):

evalnext := define_external('evalnext64s',
  cc::ARRAY(1..nn,integer[8]), 
  yy::ARRAY(1..nn,integer[8]), 
  nn::integer[4], 
  pp::integer[8],
  LIB=cat(CCodeDir, "gcd5.so") ):

evaladd := define_external('evaladd64s',
  cc::ARRAY(1..nn,integer[8]), 
  nn::integer[4], 
  TT::ARRAY(1..nn,integer[4]), 
  dx::integer[4],
  rr::ARRAY(0..dx,integer[8]), 
  pp::integer[8],
  RETURN::integer[4],
  LIB=cat(CCodeDir, "gcd5.so") ):

allocateSpace := define_external('allocateSpaceCubic',
  n::integer[4],
  dx::integer[8],
  dz::integer[8],
  du::integer[8],
  LIB=cat(CCodeDir, "hensel2.so"),
  RETURN::integer[8]):
  
BHL_C := define_external('HenselLiftCubic',
  AA::ARRAY(0..dxdz,integer[8]), dx::integer[4], dz::integer[4],
  DDX::ARRAY(1..nn,0..dy,integer[8],order=C_order), 
  DDY::ARRAY(1..nn,integer[8],order=C_order),
  F0::ARRAY(1..nn,0..du,integer[8],order=C_order), nn::integer[4], du::integer[4], 
  FF::ARRAY(1..nn,0..dxdz,integer[8],order=C_order), alpha::integer[8],
  WW::ARRAY(0..dW,integer[8],order=C_order),
  pp::integer[8], 
  LIB=cat(CCodeDir,"hensel2.so"),
  RETURN::integer[4]):

VSolve := define_external('VandermondeSolve64s',
  mm::ARRAY(1..nn,datatype=integer[8]), 
  yy::ARRAY(1..nn,datatype=integer[8]), 
  nn::integer[4],
  aa::ARRAY(1..nn,datatype=integer[8]),
  MM::ARRAY(0..nn,datatype=integer[8]),
  qq::ARRAY(0..nn,datatype=integer[8]), 
  shift::integer[4], p::integer[8], 
  LIB=cat(CCodeDir, "VSolve4.so") ):
  
#pointer := proc(f) local MaxInt := 2^63-1; addressof(f)-4*MaxInt end:

# This version evaluates into the matrix A
#EVALMOD1 := proc(AL::list,X::list,alpha::Array,AA::Array,p::integer)
#    evallistlist(pointer(AL),pointer(X),alpha,AA,p);
#end:

#BB_Maple := proc( X::list, alpha::Array, p::prime ) global AM,CNT,tBBeval,tBBdet;
#   local N := nops(X),A_eval,i,st,et,d; CNT += 1; 
#   st := time(); 
#   A_eval := Eval( AM, [seq(X[i]=alpha[i],i=1..N)]) mod p; 
#   et := time() - st; tBBeval += et; 
#   st := time();
#   d := Det(A_eval) mod p;
#   et := time() - st; tBBdet += et; 
#   return d;
#end; 

BB := proc( alpha::Array, p::prime ) global CNT, tBBdet; 
    local st,et,d,n := numelems(alpha); CNT++;
	st := time();
	d:= DetBareissSymm( alpha, n-1, p ); 
	et := time()-st; tBBdet += et; 
	return d;
end;

B_chrem := proc( B::procedure, alpha::Array, p::prime )
   local p0,d,M,d0,i,p1,d1; 
   p0 := p; d := B(alpha,p0); 
   M := p0; d0 := mods(d,M); 
   for i to 1000 do 
     p1 := nextprime(p0); 
     d1 := B(alpha,p1); 
     d := chrem([d0,d1],[M,p1]); 
     M := M*p1; d := mods(d,M);
     if d = d0 then break; 
     else d0 := d; p0 := p1; i++; 
     fi;
   od;
   printf("%d prime(s) used to interplate content\n", i);
   if i > 1000 then printf("integer content too large to interpolate\n"); return 'FAIL'; fi; 
   d;
end:

degB := proc( B::procedure, N::integer, j::posint, p::prime, tdeg::integer ) 
    # Input: B is a modular black box of a in Z[x1,...,xN], N is the no.of variables, 
    #        j is the index of a chosen degree variable, p is a prime, tdeg is a total degree bound.  
    # Output: deg(a,x_j) where a = det(A) (1<=j<=N) (w.h.p.) (with a chosen variable ordering)
    local randzp:=rand(p),h,k,beta,zeta,Z,b,hf,x;
    if j > N then printf("degB: index of the chosen degree variable out of range\n"); return 'FAIL'; fi;
    beta := Array(1..N,i->randzp());  
    Z := Array(1..tdeg+2); b := Array(1..tdeg+2);
    h[0] := 0; k := 0; 
    for k to tdeg+2 do
        do zeta := randzp(); until not member(zeta,Z); Z[k] := zeta; 
        beta[j] := zeta; 
        b[k] := B( beta, p ); 
        h[k] := Interp(Z[1..k], b[1..k], x) mod p;
    until h[k] = h[k-1];
    hf := h[k];
    if hf = 0 then return -1; else return degree(hf,x); fi;
end:

unpack := proc(f) local g; #f = factor(a1) 
    if type(f,`+`) then [[f,1]];
    elif type(f,`^`) then return [[op(f)]];
    elif type(f,`*`) then [seq( op(unpack(g)), g in f )];
    else [[f,1]];
    fi;
end:

Interp1var := proc( B::procedure, X::list, alpha::Array, deg::list, N::integer, j::posint, p::prime ) 
    # Input: The black box B, X=[x_1,...,x_N], alpha=[alpha_1,...,alpha_N], an evaluation point; 
    #        deg is the list of degrees of all variables; N is the no.of variables; 
    #        j is the chosen variable index; p is a prime.
    # Output: Det(A(alpha_1,...,x_j,...,alpha_N)) mod p
    local dj, xx, i, F, beta; 
    if nops(X)<>N then 
        printf("Interp1var: no.of variables not equal to N\n"); return 'Interp1var_FAIL'; 
    fi;
    if j > N or j < 1 then 
        printf("Interp1var: chosen variable index out of range\n"); return 'Interp1var_FAIL'; 
    fi;
    dj := deg[j]; xx := Array(1..dj+1,i->i); 
    beta := Array(alpha); F := Array(1..dj+1); 
    for i to dj+1 do beta[j] := i; F[i] := B( beta, p ); od;
    return Interp( xx, F, X[j] ) mod p;
end:

Interp1varZ_chrem := proc( B::procedure, X::list, alpha::Array, deg::list, N::integer, j::posint, p::prime )
    local p0,a0,A,M,p1,a1,i;
    p0 := p; A := Interp1var( B, X, alpha, deg, N, j, p0 ); 
    M := p0; a0 := mods(A,M); 
    for i from 1 to 2000 do
       p1 := nextprime(p0);
       a1 := Interp1var( B, X, alpha, deg, N, j, p1 ); 
       A := chrem([a0,a1],[M,p1]); 
       M := M*p1; A := mods(A,M); 
       if A = a0 then break; 
       else a0 := A; p0 := p1; i++;
       fi;
    end;
    printf("%d prime(s) used to interplate a(%a)\n",i,X[j]); 
    if i > 2000 then 
        printf("Interp1var_chrem: coefficients too large to interpolate a(%a)\n",X[j]); 
        return 'Interp1varZ_chrem_FAIL';
    fi; 
    A; 
end:

nextimage := proc(Ca,Ya,na,Ta,da,Aa,x,p)
local dega,f,i;
    evalnext(Ca,Ya,na,p);
    dega := evaladd(Ca,na,Ta,da,Aa,p);
    f := add(Aa[i]*x^i,i=0..dega);
end: 

support := proc(f, X::list) local m; 
    coeffs(f, X, 'm'); [m]; 
end:
MonCoffs := proc(f, X::list) local m; 
    return [coeffs(f, X, 'm')], [m];
end:
Tcof := proc(T, r::integer, i::integer, rr::integer ) local ca;
    if r = 1 then return Array(op(op(op(op(op(T)[2]))[2])[i])[2],datatype=integer[8]); fi;
    ca := Array(op(op(op(op(op(T)[2])[rr])[2])[i])[2],datatype=integer[8]);
  return ca;
end:
Tmon := proc(T, r::integer, i::integer, rr::integer)
    if r = 1 then return op(op(op(op(op(T)[2]))[2])[i])[1]; fi;
    op(op(op(op(op(T)[2])[rr])[2])[i])[1];
end:

MDP_Maple := proc( f0::list, ck::polynom, x, p::prime, M::list ) 
    # Input: f0, a list of r polynomials in Zp[x], ck is a polynomial in Zp[x], 
    #        p is a prime, M is a list (pre-computed polynomials) s.t. 
    #        M[r]=1,M[r-1]=f0[r],...,M[1]=f0[r]*f0[r-1]...f0[2].
    # Output: sigma, a list of r polynomials s.t.
    #        sigma[1]*f0[1]+...+sigma[r]*f0[r] = ck.
    local r:=nops(f0),MM,i,c,sigma,tau,s,t,q;
    if _params['M'] = NULL then MM[r] := 1;
        for i from r-1 by -1 to 1 do 
            MM[i] := Expand(MM[i+1]*f0[i+1]) mod p;
        end;
    else
        if r<>nops(M) then printf("MDP_Maple: sizes of f0 and M unequal\n"); 
		    return 'MDP_Maple_FAIL'; 
		fi;
        MM := M;
    fi;
    c[1] := ck;
    for i to r-1 do
  	    Gcdex(MM[i],f0[i],x,'s','t') mod p;
        sigma[i] := Rem(c[i]*s,f0[i],x,'q') mod p;
        tau[i] := Expand(c[i]*t+q*MM[i]) mod p;
        c[i+1] := tau[i];
    end;
    sigma[r]:=c[r];
    return [seq(sigma[i],i=1..r)];
end:

TaylorShift1 := proc(f,y,alpha::integer,p::prime,k::nonnegint) 
   local g,dy,A,i,r;
   if f=0 then return [] fi;
   g := f;
   if _params['k'] = NULL then 
   	  dy := degree(f,y);
   	  A := Array(0..dy);
       for i from 0 to dy do
           g := Quo(g,y-alpha,y,'r') mod p;
           A[i] := r;
       od;
       return convert(A,list);
    fi;
    for i from 0 to k do g := Quo(g,y-alpha,y,'r') mod p; od;
    return r;
end:

BHL_Maple := proc( A::polynom, F0::list, X::list, alpha::integer, p::prime ) 
    # Input: A in Zp[x,y], F0 is a list of r polynomials in Zp[x], s.t.
    #        A(y=alpha) = Lambda*product(F0[rho]) mod p, where Lambda is some constant, 
    #        X=[x,y], alpha is an evaluation point for y, p is a prime
    # Output: F, a list of r polynomials in Zp[x,y] s.t. 
    #        A = Lambda*product(F[rho]) mod p and F[rho](y=alpha) = F0[rho] for rho=1..r.
    #        Or FAIL;
    local r,x,y,NvarA,ga,gaa,a,f0,df0,dy,Mmdp,MMDP,MM,M,rr,gam,dga,ac,sigma,i,k,fprod,Dk,deltak,ck,Tf,dfSum,fbar,f,F,Fp,LCFe,eta,Fpp; 
    x := X[1]; y := X[2]; r := nops(F0); NvarA := nops(indets(A));
    if NvarA = 1 then return F0; fi; 
    if r = 1 then  
        Fp := Primpart(A,x) mod p;
        LCFe := Eval(lcoeff(Fp,x),y=alpha) mod p;
        eta := lcoeff(op(F0),x)/LCFe mod p; 
        return [eta*Fp] mod p;
    fi;
    ga := lcoeff(A, x); gaa := Eval(ga,y=alpha) mod p;
    a := Expand(ga^(r-1)*A) mod p;
    f0 := [seq(`mod`(gaa*F0[rr]/lcoeff(F0[rr],x),p), rr=1..r)];
    df0 := map(degree,f0,x); dy := degree(a,y); 
    Mmdp[r] := 1;  
    for i from r-1 by -1 to 1 do 
        Mmdp[i] := Expand(Mmdp[i+1]*f0[i+1]) mod p;
    end;
    MMDP := [seq(Mmdp[rr],rr=1..r)]; # Create a list for MDP as the last input argument
    MM := Expand(Mmdp[1]*f0[1]) mod p; # MM = product(f0[rr],rr=1..r)
    for rr to r do f[rr] := f0[rr]; 
    	    M[rr] := Quo(MM,f0[rr],x) mod p;
    od; 
    gam := TaylorShift1(ga,y,alpha,p); # A list of gamma_k's, gamma_k = coeff(ga,(y-alpha)^k)
    dga := degree(ga,y);
    ac := TaylorShift1(a,y,alpha,p);
    sigma := Array(1..r,0..dy);
    for k to dy do #printf("k = %d\n",k);
        fprod := mul(f[rr],rr=1..r) mod p; # Quintic algorithm
        Dk := TaylorShift1(fprod,y,alpha,p,k); 
        if k <= dga then 
            for rr to r do Tf[rr] := gam[k+1]*x^df0[rr];
                deltak := add(Tf[i]*M[i],i=1..r); 
            od;
        else deltak := 0;
        fi; 
        ck := Expand(ac[k+1]-Dk-deltak) mod p;
        dfSum := 0; for rr to r do dfSum := dfSum + degree(f[rr],y); od;
        if dfSum = dy and ck <> 0 then return 'BHL_Maple_FAIL(1)'; fi;
        if ck <> 0 then 
            fbar := MDP_Maple(f0,ck,x,p,MMDP); 
            for rr to r do 
            	if k <= dga then sigma[rr,k] := fbar[rr] + Tf[rr]; 
        	     else sigma[rr,k] := fbar[rr];
        	     fi;
        	     f[rr] := f[rr] + sigma[rr,k]*(y-alpha)^k;
        	  od;
        fi;  
    od;
    dfSum := 0; for rr to r do dfSum := dfSum + degree(f[rr],y); od; 
    if dfSum <> dy then return 'BHL_Maple_FAIL(2)'; fi;
    for rr to r do F[rr] := sigma[rr,k-1]; sigma[rr,0] := f0[rr]; od; 
    for i from k-1 by -1 to 1 do 
        for rr to r do F[rr] := F[rr]*(y-alpha)+sigma[rr,i-1]; od;
    od;
    for rr to r do F[rr] := Expand(F[rr]) mod p; od; 
    #printf("a-mul(F[rr])=%a",Expand(a-mul(F[rr],rr=1..r)) mod p);
    if ck <> 0 and Expand(a-mul(F[rr],rr=1..r)) mod p <> 0 then return 'BHL_Maple_FAIL(3)';
    else 
        for rr to r do 
    	    Fp[rr] := Primpart(F[rr], x) mod p;
            LCFe[rr] := Eval(lcoeff(Fp[rr],x),y=alpha) mod p; 
            eta[rr] := lcoeff(F0[rr],x)/LCFe[rr] mod p;
            Fpp[rr] := eta[rr]*Fp[rr] mod p;
        od;
        return [seq(Fpp[rr],rr=1..r)];
    fi;
end:

Initialf2matrix := proc( f0::list, x, F0::Array )
    # Input: f0 is a list of factors in Zp[x] s.t. a(y=alpha) = mul(f0[i],i=1..r) mod p, X=[x,y]
    # Ouput: Matrix F0 s.t. row rr is the array of coefficients of f0[rr] for rr=1..r
    local rr,i,r:=nops(f0),df0:=map(degree,f0,x);
    for rr to r do 
        for i from 0 to df0[rr] do F0[rr,i] := coeff(f0[rr],x,i); od; 
    od;
    return;
end:

Bivarpoly2matrix := proc( a, X::list, dx::integer, dy::integer, M::Array ) 
    # Input: a in Zp[x,y], X=[x,y], dx=deg(a,x), dy=deg(a,y)
    # Output: M with A[(dx+1)*j+i] = coeff of x^j*y^i in a, i=0..dy, j=0..dx
    local cof,mon,dmx,dmy,k,x:=X[1],y:=X[2];
    cof,mon := MonCoffs(a,X);
    dmx := map(degree,mon,x); dmy := map(degree,mon,y);
    for k to nops(mon) do M[(dx+1)*dmy[k]+dmx[k]] := cof[k]; od;
    return; 
end:

clearArray := proc( A::Array, index1::integer, index2::integer ) local i;
    for i from index1 to index2 do A[i] := 0; od;
    return; 
end:

clearArray2d := proc( A::Array, indexi1::integer, indexi2::integer, indexj1::integer, indexj2::integer ) local i,j; 
# Clear a 2-d Array A = Array(indexi1..indexi2,indexj1..indexj2)
    for i from indexi1 to indexi2 do 
        for j from indexj1 to indexj2 do A[i,j] := 0; od;
    od;
    return; 
end:

CMSHL_BBstepj := proc( fjm1::list, j::integer, X::list, alpha::Array, deg::list, N::posint, p::prime, VarPerm::nonnegint, MapleCode::list ) global CNT;
    local MapleMaxint,randzp,r,ntf,alphaa,rr,F,sigma,nsigma,degfx1,df1,M,S,s,k,test,beta,betaA,ss,f0A,dmax,di,dj,du,fA,sizeW,Wtemp,DDX,DDY,FA,TA,CA,MA,aY; 
    local Y,b,xx,II,MM,BBctr,a,f0,i,ii,jj,f,nf,Mon,Cof,mon,cof,Tf,TT,TCOF,SS,SA,mu,dx1,id,sizeVS,SSid,Mtemp,Qtemp,C,termsf,Termsf,fY,aYi,vars,nvars,univar,CAlist,MAlist,betaAm,Xm,betaAm2,varperm,x1id,betaA2,fj;
    local st0:=time(),tt2:=0,tt3:=0,tt4:=0,st,et,expandtime:=0,vstime:=0,bbtime:=0,interptime:=0,texpand,ttvs,tbb,tinterp;
    # Input: fjm1 is a list of factors in Zp[x_1,...,x_{j-1}] s.t. aj(x_j=alpha_j) = prod(fjm1[i],i=1..r),
    #        where r is the no.of factors, X=[x_1,...,x_N], alpha=[alpha_1,...,alpha_N], deg=[deg(a,x_1),...,deg(a,x_N)], p is prime
    # Output: fj = [fj_1,...,fj_r], a list of factors of aj in [x_1,...,x_j] and fj_i(x_j=alpha_j) = fjm1_i for i=1..r. Or FAIL.
    MapleMaxint:=2^63-1; r:=nops(fjm1); ntf:=map(nops, fjm1); alphaa:=p-alpha[j];
    for rr to r do
        F[rr] := addressof(fjm1[rr])-4*MapleMaxint; 
        sigma[rr] := [coeffs(fjm1[rr], X[1], 'monfx1')];
        nsigma[rr] := nops(sigma[rr]);
        degfx1[rr] := map(degree, [monfx1], X[1]);
        df1[rr] := max(degfx1[rr]); 
        M[rr] := map(support, sigma[rr], X[2..j-1]);
        s[rr] := map(nops, M[rr]);
    od;
    for k to 20 do test := 1; 
        beta := [seq(rand(p)(), i=1..j)]; Y := [seq(X[i]=beta[i], i=2..j-1)];
        for rr to r do
            S[rr] := map( Eval, M[rr], Y ) mod p;
            for i to nsigma[rr] do 
	        if nops({op(S[rr][i])}) < s[rr][i] then test := 0; next k; fi; 
	    od;
        od; 
    until test = 1;
    if k > 20 then printf( "CMSHL step %d: FAIL(1): p is likely too small\n", j ); 
	    return 'FAIL'; 
	fi;
    ss := max([seq(s[rr],rr=1..r)]); printf("CMSHL step %d: s = %a\n",j,ss);
    betaA := Array(1..j-1,beta[1..j-1],datatype=integer[8]);
    univar := Array(1..r); #indicator if a factor is univariate 
	if MapleCode[2] = 0 then
	# Set up for evaluations of fjm1
    for rr to r do
        FA[rr] := Array(0..ntf[rr],datatype=integer[8]); 
        TA[rr] := Array(0..ntf[rr],datatype=integer[4]);
        CA[rr] := Array(1..ntf[rr],datatype=integer[8]);
        MA[rr] := Array(1..ntf[rr],datatype=integer[8]);
	vars := indets(fjm1[rr]); nvars := nops(vars); 
        #printf("F[%d]=%a\n",rr,F[rr]);
        if getmapleID(F[rr])<>17 then 
            if nvars = 1 then univar[rr] := 1; k := 1; 
		        for i in degfx1[rr] do TA[rr][i] := 1; FA[rr][i] := sigma[rr][k]; k++: od; 
                for i to ntf[rr] do MA[rr][i] := 1; od;
                if df1[rr] = 1 then # fjm1[rr] is univariate and is linear
                    for i to ntf[rr] do CA[rr][i] := sigma[rr][i]; od; 
                else # fjm1[rr] is univariate but non-linear, i.e. df1[rr] > 1
                    for i to ntf[rr] do CA[rr][i] := sigma[rr][ntf[rr]-i+1]; od;
                fi;
            f0[rr] := fjm1[rr];
	        else # fjm1[rr] is linear but not univariate 
		        k := 1; 
                for i in degfx1[rr] do 
                    TA[rr][i] := s[rr][k]; 
                    FA[rr][i] := Eval(sigma[rr][k],Y) mod p; 
		            k++; 
                od;
                CAlist := map(coeffs,sigma[rr],X[2..j-1]); 
		        MAlist := map(op,S[rr]);
                for i to ntf[rr] do 
	                CA[rr][i] := CAlist[i]; 
		            MA[rr][i] := MAlist[i]; 
                od;   
	        fi;
	    else
            if nvars < j-1 then 
	            betaAm := Array(1..nvars,datatype=integer[8]); k := 1;
	            Xm := Array(1..nvars);
                    for i to j-1 do 
	                    if member(X[i],vars) then 
	                        betaAm[k] := beta[i]; Xm[k] := X[i]; k++; 
                        fi; 
		            od;
		        Xm := convert(Xm,list);
	            if VarPerm > 0 then betaAm2 := Array(1..nvars,datatype=integer[8]);
		            varperm := sort(Xm[1..nvars],output=permutation); 
                    x1id := ListTools:-Search(1,varperm); 
	                for i to nvars do betaAm2[i] := betaAm[varperm[i]]; od;
                    getsupport(F[rr],ntf[rr],betaAm2,nvars,x1id,df1[rr],TA[rr],CA[rr],MA[rr],p);
	            else getsupport(F[rr],ntf[rr],betaAm,nvars,1,df1[rr],TA[rr],CA[rr],MA[rr],p);
		        fi;
            else
		        if VarPerm > 0 then betaA2 := Array(1..j-1,datatype=integer[8]);
	                varperm := sort(X[1..j-1],output=permutation);
		            x1id := ListTools:-Search(1,varperm);
	                for i to j-1 do betaA2[i] := betaA[varperm[i]]; od;
                    getsupport(F[rr],ntf[rr],betaA2,j-1,x1id,df1[rr],TA[rr],CA[rr],MA[rr],p);
	            else getsupport(F[rr],ntf[rr],betaA,j-1,1,df1[rr],TA[rr],CA[rr],MA[rr],p);
	            fi;
            fi;
        fi; 
    od;
    fi;	 
		
    b := Array(1..N,i->1,datatype=integer[8]); 
	for i from j+1 to N do b[i] := alpha[i]; od; 
    di := deg[1]; dj := deg[j]; dmax := max(di,dj); 
    xx := Array(0..dmax, i->i+1, datatype=integer[8]); 
    II := Array(0..dmax, datatype=integer[8]); 
    NewtonInv( xx, dmax+1, II, p );
    MM := Array(0..dj,0..di,datatype=integer[8],order=C_order);

    if MapleCode[3] = 0 and r > 1 then 
        du := max(seq(df1[rr],rr=1..r)); 
        f0A := Array(1..r,0..du,datatype=integer[8],order=C_order);
        fA := Array(1..r,0..2*(di+1)*(dj+1)-1,datatype=integer[8],order=C_order);
	#sizeW := r+r*(r-1)*di+2*(di+1)*(dj+1)+9*(di+1)+4*(di+1)*(dj+1)+2*r*(di+1)*(dj+1)+5*r*(di+1)+32*(dj+1);
        sizeW := allocateSpace(r,di,dj,du);
	Wtemp := Array(1..sizeW,datatype=integer[8],order=C_order);
	DDX := Array(1..r,0..dj,datatype=integer[8],order=C_order);
        DDY := Array(1..r,datatype=integer[8],order=C_order);
    fi;

    Tf := Array(1..r); for rr to r do Tf[rr]:=table(); od;
    for k to ss do
        for i to j do b[i] := beta[i]*b[i] mod p; od;
        tbb:=time(); BBctr := CNT;
        CNT := CallBB( b, N, 1, j, di, dj, xx, MM, p, BBctr );
        bbtime+=time()-tbb;
        tinterp := time();
        Interp2var( 1, j, di, dj, dmax, xx, II, MM, p );
        interptime+=time()-tinterp;
		
		# Evaluate fjm1 at Yk = [x[2]=b[2]^k,...,x[j-1]=b[j-1]^k]
        st := time();
        if MapleCode[2] = 1 then 
            for rr to r do 
                f0[rr] := Eval(fjm1[rr],[seq(X[ii]=b[ii],ii=2..j-1)]) mod p;
               # printf("f0[%d]=%a\n",rr,f0[rr]);
            od; 
        else 
            for rr to r do 
                if univar[rr] = 0 then 
                    f0[rr] := nextimage(CA[rr],MA[rr],ntf[rr],TA[rr],df1[rr],FA[rr],X[1],p);
                fi;
               # printf("f0[%d]=%a\n",rr,f0[rr]);
                for ii from 0 to df1[rr] do f0A[rr,ii] := FA[rr][ii]; od;
            od;  
        fi;
        et := time()-st; tt2 += et;
		
        for ii to r do
            for jj from ii+1 to r do
                if Gcd(f0[ii],f0[jj]) mod p <> 1 then printf( "CMSHL step %d: FAIL(2): unlucky evaluation\n", j ); 
                    return 'FAIL'; 
                fi;
            od;
        od;
		
        st:=time(); 
        if MapleCode[3] = 1 or r = 1 then 
            a := Expand(add(add(MM[jj][ii]*X[1]^ii,ii=0..di)*X[j]^jj,jj=0..dj)) mod p;
            f := BHL_Maple( a, [seq(f0[i],i=1..r)], [X[1],X[j]], alpha[j], p ); 
	    else		
            BHL_C( MM, di, dj, DDX, DDY, f0A, r, du, fA, alpha[j], Wtemp, p ); 
            texpand := time();
	    f := [seq(add(add(fA[rr][(di+1)*jj+ii]*X[1]^ii*X[j]^jj,ii=0..DDX[rr,jj]),jj=0..DDY[rr]), rr=1..r)]; 
            expandtime += time()-texpand;
            clearArray( Wtemp,1,sizeW ); 
	    clearArray2d( f0A,1,r,0,du ); clearArray2d( fA,1,r,0,(di+1)*(dj+1)-1 ); 
	fi;
	et:=time()-st; tt3+=et;

        clearArray2d( MM,0,dj,0,di );
		
        st:=time();
        nf[k] := map( nops, f ); 
        for rr to r do TT := Tf[rr];
            Cof, Mon := MonCoffs( f[rr], [X[1],X[j]] );
            for i to nops(Mon) do
                mon := Mon[i]; cof := Cof[i]; 
                if not assigned(TT[mon]) then TT[mon] := Array(1..ss); TT[mon][k] := cof;
                else TT[mon][k] := cof;
                fi;
            od;
        od;
        et:=time()-st; tt4+=et;
    od;
	
	if MapleCode[4] = 1 then 
	    for rr to r do SS := S[rr];
            mu[rr] := max( [seq(nf[k][rr],k=1..ss)] ); 
            for i to mu[rr] do 
                dx1 := degree(Tmon(Tf,r,i,rr),X[1]); 
		id := ListTools[SearchAll](dx1,degfx1[rr]);
		sizeVS := s[rr][id]; 
		TCOF := convert(Tcof(Tf,r,i,rr)[1..sizeVS],list); 
                SSid := convert(SS[id],list); 
		ttvs:=time();
                C := VandermondeSolve( TCOF, SSid, p, 1 );
		vstime+=time()-ttvs;
                termsf[rr][i] := add(C[k]*M[rr][id][k], k=1..sizeVS);
            od;
            Termsf[rr] := Expand(add(termsf[rr][i]*Tmon(Tf,r,i,rr), i=1..mu[rr])) mod p; 
        od;
	else 
        Qtemp := Array(0..ss,datatype=integer[8]);
        for rr to r do SS := S[rr];
            mu[rr] := max( [seq(nf[k][rr],k=1..ss)] ); 
            for i to mu[rr] do
                dx1 := degree(Tmon(Tf,r,i,rr),X[1]); 
			    id := ListTools[SearchAll](dx1,degfx1[rr]); 
                sizeVS := s[rr][id]; 
			    TCOF := Tcof(Tf,r,i,rr);
                SA := Array(SS[id],datatype=integer[8]);
                Mtemp := Array(0..sizeVS, datatype=integer[8]); 
                C := Array(1..sizeVS, datatype=integer[8]); 
                #if j = N then printf( "H.L. x[%d]: rr = %d,i = %d, s[%d][%d] = %d\n", j, rr, i, rr, id, sizeVS ); fi;
                ttvs:=time();
                VSolve( SA, TCOF, sizeVS, C, Mtemp, Qtemp, 1, p );
                vstime+=time()-ttvs;
                termsf[rr][i] := add(C[k]*M[rr][id][k], k=1..s[rr][id]);
            od;
            Termsf[rr] := Expand(add(termsf[rr][i]*Tmon(Tf,r,i,rr), i=1..mu[rr])) mod p;
            #printf("Termsf[%d]=%a\n",rr,Termsf[rr]);
        od;
	fi;
    printf( "Timings: BB=%f,Interp=%f,eval=%f,BHLtot=%f,BHLexpand=%f,VStime=%f\n",bbtime,interptime,tt2,tt3,expandtime,vstime );

    # Check if a = f_1...f_r by another evaluation point 
    #b := Array(1..N,i->1,datatype=integer[8]); 
    for i to j do b[i] := rand(p)(); od;  
    Y := [seq(X[i]=b[i],i=1..j)]; 
    for rr to r do fY[rr] := Eval(Termsf[rr],Y) mod p; od; 
    aY := DetBareissSymm( b, N-1, p ); CNT++; 
    et := time()-st0;  printf( "total time for step %d  = %f\n", j, et );
    if ( aY <> modp( mul(fY[k],k=1..r), p ) ) then 
        printf( "CMSHL step %d: FAIL(3): weak SHL assumption fail\n", j ); return FAIL; 
    else fj := [seq(Termsf[rr],rr=1..r)]; #printf("f%d = %a\n",j,fj); 
        return fj;
    fi;
end:

CMSHL_BB := proc( f1::list, n::integer, X::list, alpha::Array, deg::list, N::posint, p::prime, VarPerm::nonnegint, MapleCode::list ) global CNT; 
    local a,i,r,du,di,dj,dmax,xx,II,MM,f0A,fA,sizeW,Wtemp,f,rr,ii,jj,fnm1,DDX,DDY,BBctr; 
    if n = 2 then r := nops(f1); BBctr := CNT; 
        di := deg[1]; dj := deg[n]; dmax := max(di,dj); 
        xx := Array(0..dmax, i->i+1, datatype=integer[8]); 
        II := Array(0..dmax, datatype=integer[8]); 
        NewtonInv( xx, dmax+1, II, p ); 
        MM := Array(0..dj,0..di,datatype=integer[8],order = C_order);
        CNT := CallBB( alpha, N, 1, n, deg[1], deg[n], xx, MM, p, BBctr );
        Interp2var( 1,n,di,dj,dmax,xx,II,MM,p );
        if MapleCode[3] = 1 or r = 1 then 
            a := Expand(add(add(MM[jj][ii]*X[1]^ii,ii=0..di)*X[n]^jj,jj=0..dj)) mod p;
            f := BHL_Maple( a, [seq(f1[i],i=1..r)], [X[1],X[n]], alpha[n], p ); 
	else	
            du := max(map(degree,f1,X[1]));
	    f0A := Array(1..r,0..du,datatype=integer[8],order=C_order);
            Initialf2matrix( f1, X[1], f0A );
	    fA := Array(1..r,0..2*(di+1)*(dj+1)-1,datatype=integer[8],order=C_order);
            sizeW := allocateSpace(r,di,dj,du);
	    Wtemp := Array(1..sizeW,datatype=integer[8],order=C_order); 	
            DDX := Array(1..r,0..dj,datatype=integer[8],order=C_order);
            DDY := Array(1..r,datatype=integer[8],order=C_order);			
	    BHL_C( MM, di, dj, DDX, DDY, f0A, r, du, fA, alpha[n], Wtemp, p ); 
#printf("After BHL: fA=%a\n",fA);  
            f := [seq(Expand(add(add(fA[rr][(di+1)*jj+ii]*X[1]^ii,ii=0..DDX[rr,jj])*X[n]^jj,jj=0..DDY[rr])) mod p, rr=1..r)];
#printf("After initial BHL:\nf2 = %a\n",f);
            clearArray( Wtemp,1,sizeW ); clearArray2d( f0A,1,r,0,du ); 
			clearArray2d( fA,1,r,0,(di+1)*(dj+1)-1 );        
		fi; 
		clearArray2d( MM,0,dj,0,di ); 
        return f; 
    fi; 
    fnm1 := CMSHL_BB( f1, n-1, X, alpha, deg, N, p, VarPerm, MapleCode );  
    return CMSHL_BBstepj( fnm1, n, X, alpha, deg, N, p, VarPerm, MapleCode );
end:



