
local DEBUG:
DEBUG := false:
HYUKHO := false:
TESTS := false:
CUTOFF := 32:

BUPT := proc(u,x,p) local n,f,a,m,L,R,P;
# Build the product tree
   n := nops(u);
   if n<CUTOFF then 
        f := Expand(mul(x-a,a=u)) mod p;
        return [u,f];
   fi;
   m := iquo(n,2);
   L := BUPT(u[1..m],x,p);
   R := BUPT(u[m+1..n],x,p);
   P := Expand(L[2]*R[2]) mod p;
   [u,P,L,R];
end:

DDPT := proc(T::list,f,x,p) local r,n;
# Divide down the product tree
   n := nops(T[1]);
   if n<CUTOFF then
      seq( modp(Eval(f,x=r),p), r=T[1] )
   else
      r := Rem(f,T[2],x) mod p;
      DDPT(T[3],r,x,p),DDPT(T[4],r,x,p)
   fi;
end:

FMPeval := proc(f,u,x,p) local T;
   T := BUPT(u,x,p);
   [DDPT(T,f,x,p)];
end:

if DEBUG then

   f := 3*x^3+5*x+7*x^2+1;
   u := [2,3,4,5];
   p := 31;

   T := BUPT(u,x,p);
   DDPT(T,f,x,p);
   seq(Eval(f,x=a) mod p,a=u);
   print( evalb( FMPeval(f,u,x,p)=[seq(Eval(f,x=a) mod p,a=u)] ) );

   q := 997;
   g := Randpoly(50,x) mod q;
   u := [seq(i,i=0..50)];
   #T := BUPT(u,x,q):
   #DDPT(T,g,x,q);
   print( evalb( FMPeval(g,u,x,p)=[seq(Eval(g,x=a) mod p,a=u)] ) );

fi:


FastTVS := proc(u,b,p) local x,T,M,n,D,H,Q,i,r,q,a;
# Fast transposed Vandermonde solver of Kaltofen and Yagati
   T := BUPT(u,x,p);
   M := T[2];
   n := nops(u);
   D := add(b[n-i]*x^(i+1),i=0..n-1);
   H := Expand(M*D) mod p;
   Q := add( coeff(H,x,n+i+1)*x^i,i=0..n-1 );
   q := [DDPT(T,Q,x,p)];
   r := [DDPT(T,diff(M,x) mod p,x,p)];
   a := [seq( modp(q[i]/r[i],p), i=1..n )];
end:

if DEBUG then 
   a := 2;
   n := 4;
   u := [seq(a^i mod p, i=0..n-1)];
   b := [seq(Eval(f,x=u[i]) mod p,i=1..n)];

   #trace(FastTVS);
   FastTVS(u,b,p);
   f;
fi;
   

if HYUKHO then
   p := 3*2^30+1;
   p := 116*2^55+1;
   for k from 10 to 16 do
      n := 2^k;
      f := Randpoly(n-1,x) mod p;
      alpha := 12345;
      U := Array(0..n-1);
      U[0] := alpha;
      for i to n-1 do U[i] := U[i-1]*alpha mod p; od;
      u := convert(U,list);
      b := FMPeval(f,u,x,p); 
    if k<12 then rep := 5 else rep := 1 fi;
    st := time();
      to rep do FastTVS(u,b,p); od;
    tt := (time()-st)/rep;
      printf("n=%d  time=%8.4f\n",n,tt);
  od:

fi:




# Solve [     1       1   ...     1    ] [ a1 ] = [ v1 ]
#       [    m1      m2   ...    mt    ] [ a2 ] = [ v2 ]
#       [  m1^2    m2^2   ...  mt^2    ] [ a3 ] = [ v3 ]
#       [                              ] [    ]   [    ]
#       [m1^(t-1) m2^(t-1) ... mt^(t-1)] [ at ]   [ vt ]

# The polynomial M = (x-m1)(x-m2)...(x-mt) may be input as an additional argument.
# A shift s may be input as an additional argument in which case we solve the
# shifted Vandermonde system
#
#       [   m1^s       m2^s    ...    mt^s   ] [ a1 ] = [ v1 ]
#       [ m1^(1+s)   m2^(1+s)  ...  mt^(1+s) ] [ a2 ] = [ v2 ]
#       [ m1^(2+s)   m2^(2+s)  ...  mt^(2+s) ] [ a3 ] = [ v3 ]
#       [                                    ] [    ]   [    ]
#       [m1^(t-1+s) m2^(t-1+s) ... mt^(t-1+s)] [ at ]   [ vt ]
#


VandermondeSolve := proc( v::list, m::list, p::prime, shift::integer:=0 )
local A,i,t,n;
   n := nops(v);
   if nops(m)<>n then error "wrong number of points" fi;
   A := FastTVS(m,v,p);
   if shift<>0 then
      A := Array(1..n,A);
      for i to n do
          t := 1/m[i] mod p;
          A[i] := A[i]*t &^ shift mod p;
      od;
      A := convert(A,list);
   fi;
   A;
end:


if DEBUG then
   f := x -> 23*x^4+11*x^7;
   p := 101;
   m := [ 2^4, 2^7 ];
   e := [4,7];
   t := nops(e);
   v := [ f(1), f(2) ];
   a := VandermondeSolve( v, m, p );
   add( a[i]*x^e[i], i=1..t ) - f(x);


   p := 1153;
   alpha := numtheory[primroot](p);
   e := [11,45,61,72,95,116,201,234,301,310,411,454]:
   t := nops(e);
   m := [seq( alpha &^ e[i] mod p, i=1..t )]:
   c := rand(1000):
   f := add( c()*x^e[i], i=1..t ):
   v := [seq( Eval(f,x=modp(alpha &^ i,p)) mod p, i=0..t-1 )]:
   a := VandermondeSolve( v, m, p ):
   add( a[i]*x^e[i], i=1..t ) - f;


   # Shifted example s=2
   v := [seq( Eval(f,x=modp(alpha &^ i,p)) mod p, i=2..t+1 )]:
   a := VandermondeSolve( v, m, p, 2 ):
   add( a[i]*x^e[i], i=1..t ) - f;

fi;

if TESTS then

p := 2^31-1;
#p := prevprime(2^62);
alpha := numtheory[primroot](p);
r := rand(1000000):
c := rand(p):

t := 1500;
e := {seq( r(), i=1..t )}:
while nops(e)<t do e := e union {r()} od:
m := [seq( alpha &^ e[i] mod p, i=1..t )]:
f := add( c()*x^e[i], i=1..t ):
st := time(): 
v := [seq( Eval(f,x=modp(alpha &^ i,p)) mod p, i=0..t-1 )]:
et := time()-st:
st := time(): a := VandermondeSolve( v, m, p ): mt := time()-st:
printf("t=%d  VS time=%8.3f  EVAL time=%8.3f\n",t,mt,et);
print( evalb( add( a[i]*x^e[i], i=1..t ) = f ) );

t := 3000;
e := {seq( r(), i=1..t )}:
while nops(e)<t do e := e union {r()} od:
m := [seq( alpha &^ e[i] mod p, i=1..t )]:
f := add( c()*x^e[i], i=1..t ):
st := time(): 
v := [seq( Eval(f,x=modp(alpha &^ i,p)) mod p, i=0..t-1 )]:
et := time()-st:
st := time(): a := VandermondeSolve( v, m, p ): mt := time()-st:
printf("t=%d  VS time=%8.3f  EVAL time=%8.3f\n",t,mt,et);
print( evalb( add( a[i]*x^e[i], i=1..t ) = f ) );

t := 10^4;
f := Randpoly(t-1,x) mod p:
m := [seq( alpha &^ i mod p, i=0..t-1 )]:
if nops(m) <> nops({op(m)}) then error "not unique" fi;
u := alpha*m mod p:
#st := time(): v := [seq( Eval(f,x=u[i]) mod p, i=1..t )]: et := time()-st:
st := time(): v := FMPeval(f,u,x,p): et := time()-st:
st := time(): a := VandermondeSolve( v, m, p, 1 ): mt := time()-st:
printf("t=%d  VS time=%8.3f  EVAL time=%8.3f\n",t,mt,et);
print( evalb( add( a[i]*x^(i-1), i=1..t ) = f ) );

fi:
